import unittest
from ccpi.viewer.utils import example_data


class TestExampleData(unittest.TestCase):

    def test_head_data(self):
        data = example_data.HEAD.get()
        expected_dimensions = [64, 64, 93]
        read_dimensions = data.GetDimensions()

        for i in range(0, len(expected_dimensions)):
            self.assertEqual(expected_dimensions[i], read_dimensions[i])


if __name__ == '__main__':
    unittest.main()



import unittest
from qtpy.QtWidgets import QApplication


class TestCaseQt(unittest.TestCase):

    @staticmethod
    def get_QApplication(args):
        if QApplication.instance() is None:
            return QApplication(args)
        else:
            return QApplication.instance()



# -*- coding: utf-8 -*-
#   Copyright 2017 Edoardo Pasca
#   Copyright 2018 Richard Smith
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
import vtk
from ccpi.viewer import (ALT_KEY, CONTROL_KEY, CROSSHAIR_ACTOR, CURSOR_ACTOR, HELP_ACTOR, HISTOGRAM_ACTOR,
                         LINEPLOT_ACTOR, OVERLAY_ACTOR, SHIFT_KEY, SLICE_ACTOR, SLICE_ORIENTATION_XY,
                         SLICE_ORIENTATION_XZ, SLICE_ORIENTATION_YZ)
from ccpi.viewer.CILViewerBase import CILViewerBase
from ccpi.viewer.utils import colormaps
from ccpi.viewer.utils import CameraData


class CILInteractorStyle(vtk.vtkInteractorStyleTrackballCamera):

    def __init__(self, callback):
        vtk.vtkInteractorStyleTrackballCamera.__init__(self)
        self._viewer = callback
        self.AddObserver('MouseWheelForwardEvent', self.mouseInteraction, 1.0)
        self.AddObserver('MouseWheelBackwardEvent', self.mouseInteraction, 1.0)
        self.AddObserver('KeyPressEvent', self.OnKeyPress, 1.0)
        self.AddObserver('LeftButtonPressEvent', self.OnLeftMouseClick)
        self.AddObserver('LeftButtonReleaseEvent', self.OnLeftMouseRelease)
        #self.AddObserver('RightButtonPressEvent', self.OnRightMousePress, -0.5)
        #self.AddObserver('RightButtonReleaseEvent', self.OnRightMouseRelease, -0.5)
        self.htext = None

    def GetSliceOrientation(self):
        return self._viewer.sliceOrientation

    def GetDimensions(self):
        return self._viewer.img3D.GetDimensions()

    def GetActiveSlice(self):
        return self._viewer.getActiveSlice()

    def SetActiveSlice(self, sliceno):
        self._viewer.setActiveSlice(sliceno)

    def UpdatePipeline(self, resetcamera=False):
        self._viewer.updatePipeline(resetcamera)

    def GetSliceActorNo(self):
        return self._viewer.sliceActorNo

    def SetSliceOrientation(self, orientation):
        self._viewer.sliceOrientation = orientation

    def SetActiveCamera(self, camera):
        self._viewer.ren.SetActiveCamera(camera)

    def Render(self):
        self._viewer.renWin.Render()

    def GetKeyCode(self):
        return self.GetInteractor().GetKeyCode()

    def SetKeyCode(self, keycode):
        self.GetInteractor().SetKeyCode(keycode)

    def GetControlKey(self):
        return self.GetInteractor().GetControlKey()

    def GetShiftKey(self):
        return self.GetInteractor().GetShiftKey()

    def GetAltKey(self):
        return self.GetInteractor().GetAltKey()

    def GetEventPosition(self):
        return self.GetInteractor().GetEventPosition()

    def GetActiveCamera(self):
        return self._viewer.ren.GetActiveCamera()

    def SetDecimalisation(self, value):
        decimate = self._viewer.decimate
        decimate.SetTargetReduction(value)
        if not decimate.GetInput() is None:
            decimate.Update()

    def SetEventActive(self, event):
        self._viewer.event.On(event)

    def SetEventInactive(self, event):
        self._viewer.event.Off(event)

    def GetViewerEvent(self, event):
        return self._viewer.event.isActive(event)

    def SetInitialLevel(self, level):
        self._viewer.InitialLevel = level

    def GetInitialLevel(self):
        return self._viewer.InitialLevel

    def SetInitialWindow(self, window):
        self._viewer.InitialWindow = window

    def GetInitialWindow(self):
        return self._viewer.InitialWindow

    def HideActor(self, actorno, delete=False):
        self._viewer.hideActor(actorno, delete)

    def ShowActor(self, actorno):
        self._viewer.showActor(actorno)

    def UpdateImageSlice(self):
        self._viewer.imageSlice.Update()
        self.Render()

    def mouseInteraction(self, interactor, event):
        shift = interactor.GetShiftKey()
        advance = 1
        if shift:
            advance = 10

        if event == 'MouseWheelForwardEvent':
            maxSlice = self._viewer.img3D.GetExtent()[self.GetSliceOrientation() * 2 + 1]
            # print (self.GetActiveSlice())
            if (self.GetActiveSlice() + advance <= maxSlice):
                self.SetActiveSlice(self.GetActiveSlice() + advance)
                self.UpdatePipeline()
        else:
            minSlice = self._viewer.img3D.GetExtent()[self.GetSliceOrientation() * 2]
            if (self.GetActiveSlice() - advance >= minSlice):
                self.SetActiveSlice(self.GetActiveSlice() - advance)
                self.UpdatePipeline()

    def OnLeftMouseClick(self, interactor, event):
        self.SetDecimalisation(0.8)
        self.OnLeftButtonDown()

    def OnLeftMouseRelease(self, interactor, event):
        self.SetDecimalisation(0.0)
        self.OnLeftButtonUp()

    def OnRightMousePress(self, interactor, event):
        ctrl = interactor.GetControlKey()
        alt = interactor.GetAltKey()
        shift = interactor.GetShiftKey()
        # print (alt, ctrl,shift)
        if alt and not (ctrl and shift):
            self.SetEventActive("WINDOW_LEVEL_EVENT")
        if not (alt and ctrl and shift):
            self.SetEventActive("ZOOM_EVENT")

    def OnRightMouseRelease(self, interactor, event):
        ctrl = interactor.GetControlKey()
        alt = interactor.GetAltKey()
        shift = interactor.GetShiftKey()

        # print (alt, ctrl,shift)
        if alt and not (ctrl and shift):
            self.SetEventInactive("WINDOW_LEVEL_EVENT")
        if not (alt and ctrl and shift):
            self.SetEventInactive("ZOOM_EVENT")

    def ToggleSliceInterpolation(self):
        # toggle interpolation of image slice
        is_interpolated = self._viewer.imageSlice.GetProperty().GetInterpolationType()
        if is_interpolated:
            self._viewer.imageSlice.GetProperty().SetInterpolationTypeToNearest()
        else:
            self._viewer.imageSlice.GetProperty().SetInterpolationTypeToLinear()
        self._viewer.updatePipeline()

    def SetVolumeClipping(self, clipping_on):
        if hasattr(self._viewer, 'planew') and self._viewer.clipping_plane_initialised:
            self._viewer.planew.SetEnabled(clipping_on)
            self._viewer.getRenderer().Render()
        else:
            # Doesn't exist and turn it off do nothing else:
            if clipping_on:
                planew = self.CreateClippingPlane()
                planew.On()

    def ToggleVolumeClipping(self):
        viewer = self._viewer
        viewer.imageSlice.VisibilityOff()
        # clip a volume render if available
        if hasattr(self._viewer, 'planew') and self._viewer.clipping_plane_initialised:
            is_enabled = viewer.planew.GetEnabled()
            self.SetVolumeClipping(not is_enabled)
        else:
            self.SetVolumeClipping(True)
        viewer.updatePipeline()

    def ToggleSliceVisibility(self):
        # toggle visibility of the slice
        if self._viewer.imageSlice.GetVisibility():
            self._viewer.imageSlice.VisibilityOff()
        else:
            self._viewer.imageSlice.VisibilityOn()
        self._viewer.updatePipeline()

    def SetVolumeVisibility(self, visibility):
        if not visibility:
            self._viewer.volume.VisibilityOff()
            self._viewer.light.SwitchOff()
        else:
            self._viewer.volume.VisibilityOn()
            self._viewer.light.SwitchOn()

    def ToggleVolumeVisibility(self):
        # toggle visibility of the volume render
        if not self._viewer.volume_render_initialised:
            self._viewer.installVolumeRenderActorPipeline()

        self.SetVolumeVisibility(not self._viewer.volume.GetVisibility())

        self._viewer.updatePipeline()

    def AutoWindowLevelOnVolumeRange(self, update_slice=True):
        '''Auto-adjusts window-level for the slice, based on the 5 and 95th percentiles of the whole image volume.'''
        cmin, cmax = self._viewer.getImageMapRange((5., 95.), method="scalar")
        window, level = self._viewer.getSliceWindowLevelFromRange(cmin, cmax)

        self._viewer.imageSlice.GetProperty().SetColorLevel(level)
        self._viewer.imageSlice.GetProperty().SetColorWindow(window)

        if update_slice:
            self.UpdateImageSlice()

    def OnKeyPress(self, interactor, _):
        if interactor.GetKeyCode() == "x":
            self.SetSliceOrientation(SLICE_ORIENTATION_YZ)
            self.UpdatePipeline(resetcamera=True)
        elif interactor.GetKeyCode() == "y":
            self.SetSliceOrientation(SLICE_ORIENTATION_XZ)
            self.UpdatePipeline(resetcamera=True)
        elif interactor.GetKeyCode() == "z":
            self.SetSliceOrientation(SLICE_ORIENTATION_XY)
            self.UpdatePipeline(resetcamera=True)
        elif interactor.GetKeyCode() == "a":
            self._viewer.autoWindowLevelOnSliceRange()
        elif interactor.GetKeyCode() == "h":
            self.DisplayHelp()
        elif interactor.GetKeyCode() == "r":
            filename = "current_render"
            self.SaveRender(filename)
        elif interactor.GetKeyCode() == "v":
            self.ToggleVolumeVisibility()
        elif interactor.GetKeyCode() == "s":
            self.ToggleSliceVisibility()
        elif interactor.GetKeyCode() == "i":
            self.ToggleSliceInterpolation()
        elif interactor.GetKeyCode() == "c" and self._viewer.volume_render_initialised:
            self.ToggleVolumeClipping()
        else:
            print("Unhandled event %s" % interactor.GetKeyCode())

    def CreateClippingPlane(self):
        viewer = self._viewer
        planew = vtk.vtkImplicitPlaneWidget2()

        rep = vtk.vtkImplicitPlaneRepresentation()
        extent = self.GetDataExtentInWorld()
        rep.SetWidgetBounds(*extent)
        planew.SetRepresentation(rep)
        planew.SetInteractor(viewer.getInteractor())

        rep.SetNormalToCamera()
        rep.SetOutlineTranslation(False)  # this means user can't move bounding box

        plane = vtk.vtkPlane()
        # should be in the focal point
        cam = self.GetActiveCamera()
        foc = cam.GetFocalPoint()
        plane.SetOrigin(*foc)

        proj = cam.GetDirectionOfProjection()
        proj = [x + 0.3 for x in list(proj)]
        plane.SetNormal(*proj)
        rep.SetPlane(plane)
        rep.UpdatePlacement()

        viewer.volume.GetMapper().AddClippingPlane(plane)
        viewer.volume.Modified()
        viewer.plane = plane
        viewer.planew = planew
        planew.AddObserver('InteractionEvent', self.update_clipping_plane, 0.5)
        self._viewer.clipping_plane_initialised = True

        return planew

    def update_clipping_plane(self, interactor, event):
        # event translator should you want to filter events
        # event_translator = planew.GetEventTranslator()
        # pevent = event_translator.GetTranslation(event)
        planew = self._viewer.planew
        viewer = self._viewer
        rep = planew.GetRepresentation()
        plane = vtk.vtkPlane()
        rep.GetPlane(plane)

        viewer.volume.GetMapper().RemoveAllClippingPlanes()
        viewer.volume.GetMapper().AddClippingPlane(plane)
        viewer.volume.Modified()
        viewer.getRenderer().Render()

    def DisplayHelp(self):
        help_actor = self._viewer.helpActor
        slice_actor = self._viewer.imageSlice

        if help_actor.GetVisibility():
            help_actor.VisibilityOff()
            slice_actor.VisibilityOn()
            self.ShowActor(1)
            self.Render()
            return

        font_size = 16

        # Create the text mappers and the associated Actor2Ds.

        # The font and text properties (except justification) are the same for
        # each multi line mapper. Let's create a common text property object
        multiLineTextProp = vtk.vtkTextProperty()
        multiLineTextProp.SetFontSize(font_size)
        multiLineTextProp.SetFontFamilyToArial()
        multiLineTextProp.BoldOn()
        multiLineTextProp.ItalicOn()
        multiLineTextProp.ShadowOn()
        multiLineTextProp.SetLineSpacing(1.3)

        # The text is on multiple lines and center-justified (both horizontal and
        # vertical).
        textMapperC = vtk.vtkTextMapper()
        if self.htext == None:
            self.htext = """
Mouse Interactions:
    - Slice: Mouse Scroll
    - Zoom: Right Mouse + Move Up/Down
    - Pan: Middle Mouse Button + Move or Shift + Left Mouse + Move
    - Adjust Camera: Left Mouse + Move
    - Rotate: Ctrl + Left Mouse + Move
    
Keyboard Interactions:
    h: Display this help
    x:  YZ Plane
    y:  XZ Plane
    z:  XY Plane
    r:  Save render to current_render.png
    s:  Toggle visibility of slice
    v:  Toggle visibility of volume render
    c:  Activates volume render clipping plane widget
    a:  Whole image Auto Window/Level
    """
        textMapperC.SetInput(self.htext)
        tprop = textMapperC.GetTextProperty()
        tprop.ShallowCopy(multiLineTextProp)
        tprop.SetJustificationToLeft()
        tprop.SetVerticalJustificationToCentered()
        tprop.SetColor(0, 1, 0)

        help_actor.SetMapper(textMapperC)
        help_actor.VisibilityOn()
        slice_actor.VisibilityOff()
        self.HideActor(1)

        self.Render()

    def SaveRender(self, filename):
        self._viewer.saveRender(filename)

    # Coordinate conversion ----------------------------

    def world2imageCoordinate(self, world_coordinates):
        """
        Convert from the world or global coordinates to image coordinates
        :param world_coordinates: (x,y,z)
        :return: rounded to next integer (x,y,z) in image coorindates eg. slice index
        """

        dims = self.GetInputData().GetDimensions()
        self.log(dims)
        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [round((world_coordinates[i]) / spac[i] - orig[i]) for i in range(3)]

    def world2imageCoordinateFloat(self, world_coordinates):
        """
        Convert from the world or global coordinates to image coordinates
        :param world_coordinates: (x,y,z)
        :return: float (x,y,z) in image coorindates eg. slice index
        """

        dims = self.GetInputData().GetDimensions()
        self.log(dims)
        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [(world_coordinates[i]) / spac[i] - orig[i] for i in range(3)]

    def image2world(self, image_coordinates):

        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [(image_coordinates[i]) * spac[i] + orig[i] for i in range(3)]

    def GetImageWorldExtent(self):
        """Deprecated. Use `GetDataExtentInWorld` and `GetMinMaxVoxelsFromExtent`."""
        return self.image2world(self.GetInputData().GetExtent()[1::2])

    def GetDataExtentInWorld(self):
        """
        Compute and return the extent of the input data in world coordinates.
        """
        data_extent_image = self.GetInputData().GetExtent()
        data_extent_world = self.Image2WorldExtent(data_extent_image)
        return data_extent_world

    def GetMinMaxVoxelsFromExtent(self, extent):
        """Given the extent of a box or image, gets the voxels corresponding to the min values in all directions
        and max values in all directions."""
        voxel_min = extent[0::2]
        voxel_max = extent[1::2]
        return voxel_min, voxel_max

    def Image2WorldExtent(self, extent_image):
        """Given the extent of a box or image, gets the voxels corresponding to the min values in all directions
        and max values in all directions. Then, converts their coordinates in the world coordinate system. 
        Returns the converted extent."""
        voxel_min_image, voxel_max_image = self.GetMinMaxVoxelsFromExtent(extent_image)
        voxel_min_world = self.image2world(voxel_min_image)
        voxel_max_world = self.image2world(voxel_max_image)
        extent_world = self.GetExtentFromVoxels(voxel_min_world, voxel_max_world)
        return extent_world

    def GetExtentFromVoxels(self, voxel_min, voxel_max):
        """Given the voxels corresponding to the min values in all directions
        and max values in all directions, calculates the extent of the box or image they enclose."""
        extent = (voxel_min[0], voxel_max[0], voxel_min[1], voxel_max[1], voxel_min[2], voxel_max[2])
        return extent

    def GetInputData(self):
        return self._viewer.img3D


class CILViewer(CILViewerBase):
    '''Simple 3D Viewer based on VTK classes'''

    def __init__(self, dimx=600, dimy=600, renWin=None, iren=None, ren=None, debug=False):
        CILViewerBase.__init__(self, dimx=dimx, dimy=dimy, ren=ren, renWin=renWin, iren=iren, debug=debug)
        '''creates the rendering pipeline'''

        self.setInteractorStyle(CILInteractorStyle(self))

        self.sliceActorNo = 0
        # Render decimation
        self.decimate = vtk.vtkDecimatePro()

        # Setup the slice histogram:
        self.sliceIA = vtk.vtkImageAccumulate()
        self.histogramPlotActor.SetPosition2(0.98, 0.98)
        self.histogramPlotActor.SetPosition(0., 0.)
        self.addActor(self.histogramPlotActor)
        self.histogramPlotActor.VisibilityOff()  # Off by default

        # Help text
        self.ren.AddActor(self.helpActor)

        # These may be optionally set by the user:
        self.volume_colormap_limits = None
        self.volume_colormap_name = 'viridis'

        self.volume_mapper = vtk.vtkSmartVolumeMapper()

        self.volume_render_initialised = False
        self.clipping_plane_initialised = False

    def createPolyDataActor(self, polydata):
        '''returns an actor for a given polydata'''

        self.decimate.SetInputData(polydata)
        self.decimate.SetTargetReduction(0.0)
        self.decimate.Update()

        mapper = vtk.vtkPolyDataMapper()
        if vtk.VTK_MAJOR_VERSION <= 5:
            mapper.SetInput(polydata)
        else:
            mapper.SetInputConnection(self.decimate.GetOutputPort())
        # actor
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        return actor

    def setPolyDataActor(self, actor):
        '''displays the given polydata'''

        self.hideActor(1, delete=True)
        self.ren.AddActor(actor)

        self.actors[len(self.actors) + 1] = [actor, True]
        self.iren.Initialize()
        self.renWin.Render()

    def displayPolyData(self, polydata):
        self.setPolyDataActor(self.createPolyDataActor(polydata))

    def hideActor(self, actorno, delete=False):
        '''Hides an actor identified by its number in the list of actors'''
        try:
            if self.actors[actorno][1]:
                self.ren.RemoveActor(self.actors[actorno][0])
                self.actors[actorno][1] = False

            if delete:
                self.actors = {}
                self.renWin.Render()

        except KeyError as ke:
            print("Warning Actor not present")

    def showActor(self, actorno, actor=None):
        '''Shows hidden actor identified by its number in the list of actors'''
        try:
            if not self.actors[actorno][1]:
                self.ren.AddActor(self.actors[actorno][0])
                self.actors[actorno][1] = True
                return actorno
        except KeyError as ke:
            # adds it to the actors if not there already
            if actor != None:
                self.ren.AddActor(actor)
                self.actors[len(self.actors) + 1] = [actor, True]
                return len(self.actors)

    def addActor(self, actor):
        '''Adds an actor to the render'''
        return self.showActor(0, actor)

    def setInput3DData(self, imageData):
        self.img3D = imageData

        # Have to overwrite old volume and clipping planes if they
        # were previously created:
        if self.volume_render_initialised:
            if self.clipping_plane_initialised:
                # Have to create new clipping plane so that camera
                # position is adjusted appropriately for new volume.
                # Note: just removing old plane is not sufficient.
                self.style.CreateClippingPlane()
                self.planew.SetEnabled(False)
                self.volume.GetMapper().RemoveAllClippingPlanes()
                self.clipping_plane_initialised = False

            # Have to remove old volume and install pipeline
            # with new volume:
            self.ren.RemoveVolume(self.volume)
            self.installVolumeRenderActorPipeline()

        # Reset slice visibility and orientation:
        self.imageSlice.VisibilityOn()
        self.style.SetSliceOrientation(SLICE_ORIENTATION_XY)

        # Reset camera to initial orientation
        # i.e. reset any rotation of the slice and volume
        self.resetCameraToDefault()

        # Install pipeline with new image:
        self.installPipeline()

        # needs an extra nudge to turn the slice visibility on:
        self.updatePipeline()
        # Note, this includes adjusting the camera once the new image is loaded,
        # so the camera settings have been changed.

        # Save default camera settings for this image:
        self.saveDefaultCamera()

    def setInputData(self, imageData):
        '''alias of setInput3DData'''
        return self.setInput3DData(imageData)

    def setInputAsNumpy(self, numpyarray):
        if (len(numpy.shape(numpyarray)) == 3):
            doubleImg = vtk.vtkImageData()
            shape = numpy.shape(numpyarray)
            doubleImg.SetDimensions(shape[0], shape[1], shape[2])
            doubleImg.SetOrigin(0, 0, 0)
            doubleImg.SetSpacing(1, 1, 1)
            doubleImg.SetExtent(0, shape[0] - 1, 0, shape[1] - 1, 0, shape[2] - 1)
            doubleImg.AllocateScalars(vtk.VTK_DOUBLE, 1)

            for i in range(shape[0]):
                for j in range(shape[1]):
                    for k in range(shape[2]):
                        doubleImg.SetScalarComponentFromDouble(i, j, k, 0, numpyarray[i][j][k])

            # rescale to appropriate VTK_UNSIGNED_SHORT
            stats = vtk.vtkImageAccumulate()
            stats.SetInputData(doubleImg)
            stats.Update()
            iMin = stats.GetMin()[0]
            iMax = stats.GetMax()[0]
            scale = vtk.VTK_UNSIGNED_SHORT_MAX / (iMax - iMin)

            shiftScaler = vtk.vtkImageShiftScale()
            shiftScaler.SetInputData(doubleImg)
            shiftScaler.SetScale(scale)
            shiftScaler.SetShift(iMin)
            shiftScaler.SetOutputScalarType(vtk.VTK_UNSIGNED_SHORT)
            shiftScaler.Update()
            self.img3D = shiftScaler.GetOutput()

    def installPipeline(self):
        # Reset the viewer when loading a new data source
        try:
            N = self.ren.GetActors().GetNumberOfItems()
            i = 0
            while i < N:
                actor = self.ren.GetActors().GetNextActor()
                self.ren.RemoveActor(actor)
                i += 1
        except TypeError as te:
            print(te)
            print(self.ren.GetActors())

        self.installSliceActorPipeline()
        # self.installVolumeRenderActorPipeline()

        self.ren.ResetCamera()
        self.ren.Render()

        self.adjustCamera()

        self.iren.Initialize()
        self.renWin.Render()

    def saveDefaultCamera(self):
        ''' Saves the default camera settings for a particular
        loaded 3D image.'''
        self.default_camera_data = CameraData(self.getCamera())

    def resetCameraToDefault(self):
        ''' resets to the default camera settings for the current
        loaded 3D image'''
        if hasattr(self, 'default_camera_data'):
            self.adjustCamera(resetcamera=True)
            CameraData.CopyDataToCamera(self.default_camera_data, self.getCamera())

    def setVolumeMapper(self, mapper):
        '''Sets the volume mapper to the specified mapper
        Parameters
        ----------
        mapper : vtkVolumeMapper
            The volume mapper to use, defaults is vtkSmartVolumeMapper
        '''
        self.volume_mapper = mapper

    def getVolumeMapper(self):
        '''Returns the volume mapper'''
        return self.volume_mapper

    def installVolumeRenderActorPipeline(self):
        # volume render
        volumeProperty = vtk.vtkVolumeProperty()

        self.volume_property = volumeProperty
        self.volume_mapper.SetInputData(self.img3D)

        # The volume holds the mapper and the property and
        # can be used to position/orient the volume.
        volume = vtk.vtkVolume()
        volume.SetMapper(self.volume_mapper)
        volume.SetProperty(volumeProperty)
        self.volume = volume

        # set defaults for opacity and colour mapping:
        color_percentiles = (5., 95.)
        scalar_opacity_percentiles = (80., 99.)
        gradient_opacity_percentiles = (80., 99.)
        max_opacity = 0.1

        self.setVolumeColorPercentiles(*color_percentiles, update_pipeline=False)
        self.setScalarOpacityPercentiles(*scalar_opacity_percentiles, update_pipeline=False)
        self.setGradientOpacityPercentiles(*gradient_opacity_percentiles, update_pipeline=False)
        self.setMaximumOpacity(max_opacity)

        # define colors and opacity with default values
        colors, opacity = self.getColorOpacityForVolumeRender()

        self.volume_property.SetColor(colors)

        self._setDefaultScalarOpacityFunction()

        if self.getVolumeRenderOpacityMethod() == 'scalar':
            self.volume_property.SetScalarOpacity(opacity)
        else:
            # currently this is not relevant, but in the future one may want to do
            # something fancier
            # see also https://www.kitware.com/new-in-paraview-5-9-volume-rendering-with-a-separate-opacity-array/
            self.volume_property.SetGradientOpacity(opacity)

        self.volume_property.ShadeOn()
        self.volume_property.SetInterpolationTypeToLinear()

        self.ren.AddVolume(self.volume)
        self.volume_render_initialised = True
        self.volume.VisibilityOff()
        self.addHeadlight()

    def addHeadlight(self):
        if not hasattr(self, 'light'):
            lgt = vtk.vtkLight()
            lgt.SetLightTypeToHeadlight()
            lgt.SwitchOff()
            self.getRenderer().AddLight(lgt)
            self.light = lgt

    def getVolumeRenderOpacityMethod(self):
        if not hasattr(self, '_vol_render_opacity_method'):
            self._vol_render_opacity_method = "gradient"
        return self._vol_render_opacity_method

    def setVolumeRenderOpacityMethod(self, method='gradient'):
        '''
        Parameters
        ----------
        method: string: 'scalar' or 'gradient'
            method for setting opacity of the volume render            
        '''
        if method in ['scalar', 'gradient']:
            self._vol_render_opacity_method = method
            # self.updateVolumePipeline()
            #This is a hack #TODO: fix update pipeline in case where we change opacity method
            if self.volume_render_initialised:
                planes = self.volume.GetMapper().GetClippingPlanes()
                self.ren.RemoveVolume(self.volume)
                self.volume_render_initialised = False
                self.style.ToggleVolumeVisibility()
                # add existing clipping plane to new volume
                if planes is not None:
                    plane = planes.GetItem(0)
                    self.volume.GetMapper().AddClippingPlane(plane)
                    self.volume.Modified()

    def setMaximumOpacity(self, max, update_pipeline=True):
        '''
        Parameters
        ----------
        max_opacity: float in [0,1]
            representing the maximum rendered opacity
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        self.maximum_opacity = max
        if update_pipeline:
            self.updateVolumePipeline()

    def getMaximumOpacity(self):
        '''
        Returns
        ----------
        max_opacity: float in [0,1]
            representing the maximum rendered opacity
        '''
        return self.maximum_opacity

    def setGradientOpacityPercentiles(self, min, max, update_pipeline=True):
        '''
        Parameters
        -----------
        min, max: float, default: (80., 99.)
            the percentiles on the image gradient values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'gradient'.
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        go_min, go_max = self.getImageMapRange((min, max), 'gradient')
        self.setGradientOpacityRange(go_min, go_max, update_pipeline)

    def getGradientOpacityPercentiles(self):
        '''
        Returns
        -----------
        min, max: float, default: (80., 99.)
            the percentiles on the image gradient values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'gradient'.
        '''
        go_min, go_max = self.getGradientOpacityRange()
        value_min, value_max = self.getImageMapWholeRange('gradient')
        min_percentage = (go_min - value_min) / (value_max - value_min) * 100
        max_percentage = (go_max - value_min) / (value_max - value_min) * 100
        return min_percentage, max_percentage

    def setScalarOpacityPercentiles(self, min, max, update_pipeline=True):
        '''
        min, max: float, default: (80., 99.)
            the percentiles on the image values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'scalar'.
        '''
        so_min, so_max = self.getImageMapRange((min, max), 'scalar')
        self.setScalarOpacityRange(so_min, so_max, update_pipeline)

    def getScalarOpacityPercentiles(self):
        '''
        Returns
        -----------
        min, max: float, default: (80., 99.)
            the percentiles on the image values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'scalar'.
        '''
        so_min, so_max = self.getScalarOpacityRange()
        value_min, value_max = self.getImageMapWholeRange('scalar')
        min_percentage = (so_min - value_min) / (value_max - value_min) * 100
        max_percentage = (so_max - value_min) / (value_max - value_min) * 100
        return min_percentage, max_percentage

    def setVolumeColorPercentiles(self, min, max, update_pipeline=True):
        '''
        Parameters
        -----------
        min, max: int, default: (85., 95.)
            the percentiles on the image values upon which the colours will be mapped to
        '''
        cmin, cmax = self.getImageMapRange((min, max), 'scalar')
        self.setVolumeColorRange(cmin, cmax, update_pipeline)

    def getVolumeColorPercentiles(self):
        '''
        Returns
        -----------
        min, max: int, default: (85., 95.)
            the percentiles on the image values upon which the colours will be mapped to
        '''
        cmin, cmax = self.getVolumeColorRange()
        value_min, value_max = self.getImageMapWholeRange('scalar')
        min_percentage = (cmin - value_min) / (value_max - value_min) * 100
        max_percentage = (cmax - value_min) / (value_max - value_min) * 100
        return min_percentage, max_percentage

    def setGradientOpacityRange(self, min, max, update_pipeline=True):
        '''
        Parameters
        -----------
        min, max: float, default: (80., 99.)
            the upper and lower image gradient values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'gradient'.
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        self.gradient_opacity_limits = (min, max)
        if update_pipeline:
            self.updateVolumePipeline()

    def getGradientOpacityRange(self):
        '''
        Returns
        -----------
        (min, max): tuple, default: (80., 99.)
            the upper and lower image gradient values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'gradient'.
        '''
        return self.gradient_opacity_limits

    def setScalarOpacityRange(self, min, max, update_pipeline=True):
        '''
        Parameters
        -----------
        min, max: float, default: (80., 99.)
            the upper and lower image values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'scalar'.
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        self.scalar_opacity_limits = (min, max)
        if update_pipeline:
            self.updateVolumePipeline()

    def getScalarOpacityRange(self):
        '''
        Returns
        -----------
        (min, max): tuple, default: (80., 99.)
            the upper and lower image values that the 
            opacity will be mapped to if setVolumeRenderOpacityMethod
            has been set to 'scalar'.
        '''
        return self.scalar_opacity_limits

    def setVolumeColorRange(self, min, max, update_pipeline=True):
        '''
        Parameters
        -----------
        min, max: float, default: the raw value of the 80. percentile for min, and the raw value of the 99. percentile for max.
            the upper and lower image values that the 
            color will be mapped to.
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        self.volume_colormap_limits = (min, max)
        if update_pipeline:
            self.updateVolumePipeline()

    def getVolumeColorRange(self):
        '''
        Returns
        -----------
        (min, max): tuple, default: (80., 99.)
            the upper and lower image values that the 
            color will be mapped to.
        '''
        return self.volume_colormap_limits

    def setVolumeColorMapName(self, cmap='viridis'):
        '''set the volume color map name
        Parameters
        ----------
        cmap: string, default: 'viridis'
            with one of ['viridis', 'plasma', 'magma', 'inferno'],
            or matplotlib's cmaps if available
        '''
        self.volume_colormap_name = cmap
        self.updateVolumePipeline()

    def getVolumeColorMapName(self):
        '''get the volume color map name'''
        return self.volume_colormap_name

    def _setDefaultScalarOpacityFunction(self):
        # used inside viewer, not for user
        self.default_scalar_opacity = self.volume_property.GetScalarOpacity()

    def _getDefaultScalarOpacityFunction(self):
        # used inside viewer, not for user
        return self.default_scalar_opacity

    def getColorOpacityForVolumeRender(self, color_num=255):
        '''
        Defines the color and opacity tables
        
        Parameters
        ----------
        color_num: int, default: 255 
            number of colors in the map
        '''

        colors = colormaps.CILColorMaps.get_color_transfer_function(self.getVolumeColorMapName(),
                                                                    self.volume_colormap_limits)

        method = self.getVolumeRenderOpacityMethod()

        if method == 'scalar':
            omin, omax = self.scalar_opacity_limits
        else:
            omin, omax = self.gradient_opacity_limits

        # mapping values in the image or gradient to the opacity:
        x = self.getMappingArray(color_num, method)
        opacity = colormaps.CILColorMaps.get_opacity_transfer_function(x, colormaps.relu, omin, omax,
                                                                       self.maximum_opacity)

        return colors, opacity

    def getMappingArray(self, color_num, method):
        '''
        generates array of color_num values between min and max values in 
        image or image gradient (depending on method).
        '''
        ia = self.getImageHistogramStatistics(method)
        x = numpy.linspace(ia.GetMinimum(), ia.GetMaximum(), num=color_num)
        return x

    def installSliceActorPipeline(self):
        self.voi.SetInputData(self.img3D)

        extent = [i for i in self.img3D.GetExtent()]
        for i in range(len(self.slicenos)):
            self.slicenos[i] = round((extent[i * 2 + 1] + extent[i * 2]) / 2)
        extent[self.sliceOrientation * 2] = self.getActiveSlice()
        extent[self.sliceOrientation * 2 + 1] = self.getActiveSlice()

        self.voi.SetVOI(extent[0], extent[1], extent[2], extent[3], extent[4], extent[5])

        self.voi.Update()

        self.ia.SetInputData(self.voi.GetOutput())
        self.ia.Update()

        self.style.AutoWindowLevelOnVolumeRange(update_slice=False)
        self.InitialLevel = self.getSliceColorLevel()
        self.InitialWindow = self.getSliceColorWindow()

        self.imageSliceMapper.SetInputConnection(self.voi.GetOutputPort())
        self.imageSlice.Update()

        self.imageSlice.GetProperty().SetColorLevel(self.InitialLevel)
        self.imageSlice.GetProperty().SetColorWindow(self.InitialWindow)
        self.imageSlice.GetProperty().SetInterpolationTypeToNearest()
        self.imageSlice.GetProperty().SetOpacity(0.99)

        self.imageSlice.Update()

        self.ren.AddActor(self.imageSlice)

    def updatePipeline(self, resetcamera=False):
        self.hideActor(self.sliceActorNo)

        extent = [i for i in self.img3D.GetExtent()]
        extent[self.sliceOrientation * 2] = self.getActiveSlice()
        extent[self.sliceOrientation * 2 + 1] = self.getActiveSlice()
        self.voi.SetVOI(extent[0], extent[1], extent[2], extent[3], extent[4], extent[5])

        self.voi.Update()
        self.ia.Update()

        self.imageSliceMapper.SetOrientation(self.sliceOrientation)
        self.imageSlice.Update()

        no = self.showActor(self.sliceActorNo, self.imageSlice)
        self.sliceActorNo = no

        self.updateVolumePipeline()
        self.updateSliceHistogram()

        self.adjustCamera(resetcamera)

        self.renWin.Render()

    def updateVolumePipeline(self):
        if self.volume_render_initialised and self.volume.GetVisibility():
            # define colors and opacity with default values
            colors, opacity = self.getColorOpacityForVolumeRender()

            self.volume_property.SetColor(colors)

            # Update whether we use our calculated opacity as the scalar or gradient opacity
            if self.getVolumeRenderOpacityMethod() == 'gradient':
                # Also return the scalar opacity to its default value:
                # If we don't do this then the gradient opacity changes depending on what the
                # user set for the scalar opacity - not sure we want this:
                self.volume_property.SetScalarOpacity(self._getDefaultScalarOpacityFunction())
                self.volume_property.DisableGradientOpacityOff()
                self.volume_property.SetGradientOpacity(opacity)

            elif self.getVolumeRenderOpacityMethod() == 'scalar':
                self.volume_property.DisableGradientOpacityOn()
                self.volume_property.SetScalarOpacity(opacity)

            self.renWin.Render()

    def adjustCamera(self, resetcamera=False):
        self.ren.ResetCameraClippingRange()

        if resetcamera:
            self.ren.ResetCamera()

    def updateSliceHistogram(self):
        irange = self.voi.GetOutput().GetScalarRange()
        self.sliceIA.SetInputData(self.voi.GetOutput())
        self.sliceIA.IgnoreZeroOn()

        #use 255 bins
        delta = irange[1] - irange[0]
        nbins = 255
        self.sliceIA.SetComponentSpacing(delta / nbins, 0, 0)
        self.sliceIA.SetComponentExtent(0, nbins - 1, 0, 0, 0, 0)
        self.sliceIA.Update()

        self.histogramPlotActor.AddDataSetInputConnection(self.sliceIA.GetOutputPort())
        self.histogramPlotActor.SetXRange(irange[0], irange[1])
        self.histogramPlotActor.SetYRange(self.sliceIA.GetOutput().GetScalarRange())

    def remove_clipping_plane(self):
        if self.volume_render_initialised and self.clipping_plane_initialised:
            self.volume.GetMapper().RemoveAllClippingPlanes()

            # Now remove planew from the cil_viewer
            del self.planew
            del self.plane
            self.clipping_plane_initialised = False

            self.getRenderer().Render()
            self.updatePipeline()



import sys
from functools import partial

import vtk
from ccpi.viewer import viewer2D, viewer3D
from ccpi.viewer.ui.main_windows import TwoViewersMainWindow
from qtpy import QtWidgets
from qtpy.QtWidgets import QAction, QCheckBox


class StandaloneViewerMainWindow(TwoViewersMainWindow):
    '''
    A main window for displaying two viewers side by side, with a menu bar
    for selecting the images to be displayed.

    The first image selected will be displayed on both viewers. The second
    is an image overlay which is displayed on the 2D viewer/s only.

    A dock widget is created which contains widgets for displaying the
    file name of the image shown on the viewer, and the level of downsampling
    of the image displayed on the viewer.    
    '''

    def __init__(self,
                 title="StandaloneViewer",
                 app_name="Standalone Viewer",
                 settings_name=None,
                 organisation_name=None,
                 viewer1_type='2D',
                 viewer2_type='3D'):
        super(StandaloneViewerMainWindow, self).__init__(title, app_name, settings_name, organisation_name,
                                                         viewer1_type, viewer2_type)

        self.addToMenu()
        self.addToViewerCoordsDockWidget()

        self.image_overlay = vtk.vtkImageData()

    def addToMenu(self):
        '''
        Adds actions to the menu bar for selecting the images to be displayed
        '''
        file_menu = self.menus['File']

        # insert image selection as first action in file menu:

        image2_action = QAction("Select Image Overlay", self)
        image2_action.triggered.connect(lambda: self.setViewersInputFromDialog(self.viewers, input_num=2))
        file_menu.insertAction(file_menu.actions()[0], image2_action)

        image1_action = QAction("Select Image", self)
        image1_action.triggered.connect(lambda: self.setViewersInputFromDialog(self.viewers))
        file_menu.insertAction(file_menu.actions()[0], image1_action)

    def addToViewerCoordsDockWidget(self):
        '''
        Adds widgets to the viewer coords dock widget for displaying the
        image overlay, and for showing/hiding the 2D and 3D viewers.
        '''
        checkbox = QCheckBox("Show Image Overlay")
        checkbox.setChecked(True)
        self.viewer_coords_dock.widget().addSpanningWidget(checkbox, 'image_overlay')
        checkbox.stateChanged.connect(partial(self.showHideImageOverlay))

        checkbox2 = QCheckBox("Show 2D Viewer")
        checkbox2.setChecked(True)
        checkbox2.stateChanged.connect(partial(self.showHideViewer, 0))

        checkbox3 = QCheckBox("Show 3D Viewer")
        checkbox3.setChecked(True)
        self.viewer_coords_dock.widget().addWidget(checkbox3, checkbox2, 'show_viewer')
        checkbox3.stateChanged.connect(partial(self.showHideViewer, 1))

    def showHideImageOverlay(self, show=True):
        '''
        Shows or hides the image overlay/s on the viewers
        '''
        for viewer in self.viewer_coords_dock.viewers:
            if isinstance(viewer, viewer2D):
                if show:
                    if hasattr(self, 'image_overlay'):
                        viewer.setInputData2(self.image_overlay)
                else:
                    self.image_overlay = viewer.image2
                    viewer.setInputData2(vtk.vtkImageData())


class standalone_viewer(object):
    '''
    Launches a StandaloneViewerMainWindow instance.

    Parameters:
    ------------
    title: str
        title of the window
    viewer1_type: '2D' or '3D'
    viewer2_type: '2D', '3D' or None 
        if None, only one viewer is displayed
    '''

    def __init__(self, title="", viewer1_type='2D', viewer2_type='3D', *args, **kwargs):
        '''Creator
        
        Parameters:
        ------------
        title: str
            title of the window
        viewer1_type: '2D' or '3D'
        viewer2_type: '2D', '3D' or None 
            if None, only one viewer is displayed
        '''
        app = QtWidgets.QApplication(sys.argv)
        self.app = app

        self.set_up(title, viewer1_type, viewer2_type, *args, **kwargs)

    def set_up(self, title, viewer1_type, viewer2_type=None, *args, **kwargs):
        '''
        Sets up the standalone viewer.

        Parameters:
        ------------
        title: str
            title of the window
        viewer1_type: '2D' or '3D'
        viewer2_type: '2D', '3D' or None 
            if None, only one viewer is displayed
        '''

        window = StandaloneViewerMainWindow(title, viewer1_type, viewer2_type)

        self.window = window
        self.has_run = None

        window.show()

    def show(self):
        '''
        Shows the window
        '''
        if self.has_run is None:
            self.has_run = self.app.exec_()
        else:
            print('No instance can be run interactively again. Delete and re-instantiate.')

    def __del__(self):
        '''destructor'''
        self.app.exit()


def main():
    # Run a standalone viewer with a 2D and a 3D viewer:
    err = vtk.vtkFileOutputWindow()
    err.SetFileName("viewer.log")
    vtk.vtkOutputWindow.SetInstance(err)
    standalone_viewer_instance = standalone_viewer("Standalone Viewer", viewer1_type='2D', viewer2_type='3D')
    standalone_viewer_instance.show()
    return 0


if __name__ == '__main__':
    main()


import vtk
from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase
import h5py
import numpy as np
from vtk.numpy_interface import dataset_adapter as dsa
from vtk.util import numpy_support

# Methods for reading and writing HDF5 files:


def write_image_data_to_hdf5(filename, data, dataset_name, attributes={}):
    '''
    Writes vtkImageData to a dataset in a HDF5 file

    Args:
        filename: string - name of HDF5 file to write to.
        data: vtkImageData - image data to write.
        DatasetName: string - DatasetName for HDF5 dataset.
        attributes: dict - attributes to assign to HDF5 dataset.
    '''

    with h5py.File(filename, "a") as f:
        # The function imgdata.GetPointData().GetScalars() returns a pointer to a
        # vtk<TYPE>Array where the data is stored as X-Y-Z.
        array = numpy_support.vtk_to_numpy(data.GetPointData().GetScalars())

        # Note that we flip the dimensions here because
        # VTK's order is Fortran whereas h5py writes in
        # C order. We don't want to do deep copies so we write
        # with dimensions flipped and pretend the array is
        # C order.
        array = array.reshape(data.GetDimensions()[::-1])
        try:
            dset = f.create_dataset(dataset_name, data=array)
        except RuntimeError:
            print("Unable to save image data to {0}."
                  "Dataset with name {1} already exists in this file.".format(filename, dataset_name))
            return
        for key, value in attributes.items():
            dset.attrs[key] = value


class HDF5Reader(VTKPythonAlgorithmBase):
    '''
    vtkAlgorithm for reading vtkImageData from a HDF5 file
    Adapted from:
    https://blog.kitware.com/developing-hdf5-readers-using-vtkpythonalgorithm/
    To use this, you must set a FileName (the file you are reading), and also a
    DatasetName (the name of where in the hdf5 file the data will be saved)
    '''

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=1, outputType='vtkImageData')

        self._FileName = None
        self._DatasetName = None
        self._4DSliceIndex = 0
        self._4DIndex = 0

    def RequestData(self, request, inInfo, outInfo):
        self._update_output_data(outInfo)
        return 1

    def _update_output_data(self, outInfo):
        if self._DatasetName is None:
            raise Exception("DataSetName must be set.")
        if self._FileName is None:
            raise Exception("FileName must be set.")
        with h5py.File(self._FileName, 'r') as f:
            info = outInfo.GetInformationObject(0)
            shape = np.shape(f[self._DatasetName])
            # print("keys:", list(f.keys()))
            # print(shape)
            ue = info.Get(vtk.vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT())
            # Note that we flip the update extents because VTK is Fortran order
            # whereas h5py reads in C order. When writing we pretend that the
            # data was C order so we have to flip the extents/dimensions.
            if len(shape) == 3:
                data = f[self._DatasetName][ue[4]:ue[5] + 1, ue[2]:ue[3] + 1, ue[0]:ue[1] + 1]
            elif len(shape) == 4:
                if self._4DIndex == 0:
                    data = f[self._DatasetName][self._4DSliceIndex, ue[4]:ue[5] + 1, ue[2]:ue[3] + 1, ue[0]:ue[1] + 1]
                elif self._4DIndex == 1:
                    data = f[self._DatasetName][ue[4]:ue[5] + 1, self._4DSliceIndex, ue[2]:ue[3] + 1, ue[0]:ue[1] + 1]
                elif self._4DIndex == 2:
                    data = f[self._DatasetName][ue[4]:ue[5] + 1, ue[2]:ue[3] + 1, self._4DSliceIndex, ue[0]:ue[1] + 1]
                elif self._4DIndex == 3:
                    data = f[self._DatasetName][ue[4]:ue[5] + 1, ue[2]:ue[3] + 1, ue[0]:ue[1] + 1, self._4DSliceIndex]
            else:
                raise Exception("Currently only 3D and 4D datasets are supported.")
            # print("attributes: ", f.attrs.items())
            output = dsa.WrapDataObject(vtk.vtkImageData.GetData(outInfo))
            output.SetExtent(ue)
            output.PointData.append(data.ravel(), self._DatasetName)
            output.PointData.SetActiveScalars(self._DatasetName)
            return output

    def SetFileName(self, fname):
        if fname != self._FileName:
            self.Modified()
            if self._DatasetName is not None:
                with h5py.File(fname, 'r') as f:
                    if not (self._DatasetName in f):
                        raise Exception("No dataset named {} exists in {}.".format(self._DatasetName, fname))
            self._FileName = fname

    def GetFileName(self):
        return self._FileName

    def SetDatasetName(self, lname):
        if lname != self._DatasetName:
            self.Modified()
            if self._FileName is not None:
                with h5py.File(self._FileName, 'r') as f:
                    if not (lname in f):
                        raise Exception("No dataset named {} exists in {}.".format(lname, self._FileName))
            self._DatasetName = lname

    def GetDatasetName(self):
        return self._DatasetName

    def Set4DIndex(self, index):
        '''Sets which index is the 4th dimension that we will only read 1 slice of.'''
        if index not in range(0, 4):
            raise Exception("4D Index must be between 0 and 3.")
        if index != self._4DIndex:
            self.Modified()
            self._4DIndex = index

    def Set4DSliceIndex(self, index):
        '''Sets which index to read along the 4th dimension.'''
        if index != self._4DSliceIndex:
            self.Modified()
            self._4DSliceIndex = index

    def GetDimensions(self):
        if self._FileName is None:
            raise Exception("FileName must be set.")
        with h5py.File(self._FileName, 'r') as f:
            # Note that we flip the shape because VTK is Fortran order
            # whereas h5py reads in C order. When writing we pretend that the
            # data was C order so we have to flip the extents/dimensions.
            if self._DatasetName is None:
                raise Exception("DataSetName must be set.")
            return f[self._DatasetName].shape[::-1]

    def GetDataSetAttributes(self):
        if self._FileName is None:
            raise Exception("FileName must be set.")
        with h5py.File(self._FileName, 'r') as f:
            if self._DatasetName is None:
                raise Exception("DataSetName must be set.")
            return dict(f[self._DatasetName].attrs)

    def GetOrigin(self):
        # There is not a standard way to set the origin in a HDF5
        # file so we do not have a way to read it. Therefore we
        # assume it is at 0,0,0
        return (0, 0, 0)

    def GetDataType(self):
        if self._FileName is None:
            raise Exception("FileName must be set.")
        with h5py.File(self._FileName, 'r') as f:
            data_type = f.get(self._DatasetName).dtype
            return data_type

    def RequestInformation(self, request, inInfo, outInfo):
        dims = self.GetDimensions()
        info = outInfo.GetInformationObject(0)
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), (0, dims[0] - 1, 0, dims[1] - 1, 0, dims[2] - 1),
                 6)
        return 1


class HDF5SubsetReader(VTKPythonAlgorithmBase):
    '''Modifies a HDF5Reader to return a different extent from an HDF5 file
    
    
    Examples:
    ---------
    
    reader = HDF5Reader() 
    reader.SetFileName('file.h5') 
    reader.SetDatasetName("ImageData") 
    
    cropped_reader = HDF5SubsetReader() 
    cropped_reader.SetInputConnection(reader.GetOutputPort()) 
    cropped_reader.SetUpdateExtent((0, 2, 3, 5, 1, 2)) 
    cropped_reader.Update() 
    '''

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self,
                                        nInputPorts=1,
                                        inputType='vtkImageData',
                                        nOutputPorts=1,
                                        outputType='vtkImageData')
        self.__UpdateExtent = None

    def RequestInformation(self, request, inInfo, outInfo):
        info = outInfo.GetInformationObject(0)
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), self.__UpdateExtent, 6)
        return 1

    def RequestUpdateExtent(self, request, inInfo, outInfo):
        if self.__UpdateExtent is not None:
            info = inInfo[0].GetInformationObject(0)

            whole_extent = info.Get(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT())
            set_extent = list(info.Get(vtk.vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT()))

            for i, value in enumerate(set_extent):
                if value == -1:
                    set_extent[i] = whole_extent[i]
                else:
                    if i % 2 == 0:
                        if value < whole_extent[i]:
                            raise ValueError("Requested extent {}\
                                is outside of original image extent {} as {}<{}".format(
                                set_extent, whole_extent, value, whole_extent[i]))
                    else:
                        if value > whole_extent[i]:
                            raise ValueError("Requested extent {}\
                                is outside of original image extent {} as {}>{}".format(
                                set_extent, whole_extent, value, whole_extent[i]))

            self.SetUpdateExtent(set_extent)

            info.Set(vtk.vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT(), self.__UpdateExtent, 6)
        return 1

    def RequestData(self, request, inInfo, outInfo):
        inp = vtk.vtkImageData.GetData(inInfo[0])
        opt = vtk.vtkImageData.GetData(outInfo)
        opt.ShallowCopy(inp)
        return 1

    def SetUpdateExtent(self, ue):
        if ue != self.__UpdateExtent:
            self.Modified()
            self.__UpdateExtent = ue

    def GetUpdateExtent(self):
        return self.__UpdateExtent

    def GetOutput(self):
        return self.GetOutputDataObject(0)



from eqt.ui.UIFormWidget import FormDockWidget
from qtpy.QtWidgets import QComboBox, QLabel
from ccpi.viewer.CILViewer import CILViewer
from ccpi.viewer.CILViewer2D import CILViewer2D


class ViewerCoordsDockWidget(FormDockWidget):
    ''' This is the dockwidget which
    shows the original and downsampled image
    size and the user can select whether coordinates
    are displayed in system of original or downsampled image'''

    def __init__(self, parent, title="Viewer Information", viewers=None):
        '''
        Parameters
        ----------
        parent : QWidget
            The parent widget.
        viewers : list of CILViewer2D and/or CILViewer instances, default = None
            The viewers which this dock widget will display information for.
        '''
        super(ViewerCoordsDockWidget, self).__init__(parent, title)

        self.setViewers(viewers)

        form = self.widget()

        viewer_coords_widgets = {}

        viewer_coords_widgets['image'] = QLabel()
        viewer_coords_widgets['image'].setText("Display image: ")

        viewer_coords_widgets['image_combobox'] = QComboBox()
        viewer_coords_widgets['image_combobox'].setEnabled(False)
        form.addWidget(viewer_coords_widgets['image_combobox'], viewer_coords_widgets['image'], 'image')

        viewer_coords_widgets['coords_info'] = QLabel()
        viewer_coords_widgets['coords_info'].setText(
            "The viewer displays a downsampled image for visualisation purposes: ")
        viewer_coords_widgets['coords_info'].setVisible(False)

        form.addSpanningWidget(viewer_coords_widgets['coords_info'], 'coords_info')

        form.addWidget(QLabel(""), QLabel("Loaded Image Size: "), 'loaded_image_dims')

        viewer_coords_widgets['displayed_image_dims_label'] = QLabel()
        viewer_coords_widgets['displayed_image_dims_label'].setText("Displayed Image Size: ")
        viewer_coords_widgets['displayed_image_dims_label'].setVisible(False)

        viewer_coords_widgets['displayed_image_dims_field'] = QLabel()
        viewer_coords_widgets['displayed_image_dims_field'].setText("")
        viewer_coords_widgets['displayed_image_dims_field'].setVisible(False)

        form.addWidget(viewer_coords_widgets['displayed_image_dims_field'],
                       viewer_coords_widgets['displayed_image_dims_label'], 'displayed_image_dims')

        viewer_coords_widgets['coords'] = QLabel()
        viewer_coords_widgets['coords'].setText("Display viewer coordinates in: ")

        viewer_coords_widgets['coords_combo'] = QComboBox()
        viewer_coords_widgets['coords_combo'].addItems(["Loaded Image", "Downsampled Image"])
        viewer_coords_widgets['coords_combo'].setEnabled(False)
        form.addWidget(viewer_coords_widgets['coords_combo'], viewer_coords_widgets['coords'], 'coords_combo')

        viewer_coords_widgets['coords_warning'] = QLabel()
        viewer_coords_widgets['coords_warning'].setText("Warning: These coordinates are approximate.")
        viewer_coords_widgets['coords_warning'].setVisible(False)

        form.addSpanningWidget(viewer_coords_widgets['coords_warning'], 'coords_warning')

    @property
    def viewers(self):
        ''' Get the viewers which this dock widget will display information for.
        
        Returns
        -------
        list of CILViewer2D and/or CILViewer3D instances
            The viewers which this dock widget will display information for.'''
        return self._viewers

    @viewers.setter
    def viewers(self, viewers):
        '''
        Set the viewers which this dock widget will display information for.

        Parameters
        ----------
        viewers : list of CILViewer2D and/or CILViewer instances
            The viewers which this dock widget will display information for.
        '''
        if viewers is not None:
            if not isinstance(viewers, list):
                viewers = [viewers]
            for viewer in viewers:
                if not isinstance(viewer, (CILViewer2D, CILViewer)):
                    raise TypeError("viewers must be a list of CILViewer2D and/or CILViewer instances")
        self._viewers = viewers

    def setViewers(self, viewers):
        ''' Set the viewers which this dock widget will display information for.

        Parameters
        ----------
        viewers : list of CILViewer2D and/or CILViewer instances
            The viewers which this dock widget will display information for.
        '''
        self.viewers = viewers

    def getViewers(self):
        ''' Get the viewers which this dock widget will display information for.
        
        Returns
        -------
        list of CILViewer2D and/or CILViewer instances
            The viewers which this dock widget will display information for.'''
        return self.viewers



import vtk
import sys
import vtk
from qtpy import QtCore, QtWidgets
from ccpi.viewer.QCILRenderWindowInteractor import QCILRenderWindowInteractor
from ccpi.viewer import viewer2D, viewer3D


class QCILViewerWidget(QtWidgets.QFrame):
    '''A QFrame to embed in Qt application containing a VTK Render Window
    
    All the interaction is passed from Qt to VTK.

    :param viewer: The viewer you want to embed in Qt: CILViewer2D or CILViewer
    :param interactorStyle: The interactor style for the Viewer. 
    '''

    def __init__(self,
                 parent,
                 viewer,
                 shape=(600, 600),
                 debug=False,
                 renderer=None,
                 interactorStyle=None,
                 enableSliderWidget=True):
        '''Creator. Creates an instance of a QFrame and of a CILViewer
        
        The viewer is placed in the QFrame inside a QVBoxLayout. 
        The viewer is accessible as member 'viewer'
        '''

        super(QCILViewerWidget, self).__init__(parent=parent)
        # currently the size of the frame is set by stretching to the whole
        # area in the main window. A resize of the MainWindow triggers a resize of
        # the QFrame to occupy the whole area available.

        dimx, dimy = shape
        # self.resize(dimx, dimy)

        self.vtkWidget = QCILRenderWindowInteractor(self)

        if renderer is None:
            self.ren = vtk.vtkRenderer()
        else:
            self.ren = renderer

        self.vtkWidget.GetRenderWindow().AddRenderer(self.ren)
        # https://discourse.vtk.org/t/qvtkwidget-render-window-is-outside-main-qt-app-window/1539/8?u=edoardo_pasca
        self.iren = self.vtkWidget.GetRenderWindow().GetInteractor()

        if viewer is viewer2D:
            self.viewer = viewer(dimx=dimx,
                                 dimy=dimy,
                                 ren=self.ren,
                                 renWin=self.vtkWidget.GetRenderWindow(),
                                 iren=self.iren,
                                 debug=debug,
                                 enableSliderWidget=enableSliderWidget)
            al = self.viewer.axisLabelsText
            self.viewer.setAxisLabels([al[0], al[1], ''], False)
        elif viewer is viewer3D:
            self.viewer = viewer(dimx=dimx,
                                 dimy=dimy,
                                 ren=self.ren,
                                 renWin=self.vtkWidget.GetRenderWindow(),
                                 iren=self.iren,
                                 debug=debug)
        else:
            raise KeyError("Viewer class not provided. Submit an uninstantiated viewer class object"
                           "using 'viewer' keyword")

        if interactorStyle is not None:
            self.viewer.style = interactorStyle(self.viewer)
            self.viewer.iren.SetInteractorStyle(self.viewer.style)

        self.vl = QtWidgets.QVBoxLayout()
        self.vl.addWidget(self.vtkWidget)
        self.setLayout(self.vl)
        self.adjustSize()


class QCILDockableWidget(QtWidgets.QDockWidget):
    '''Inserts a vtk viewer in a dock widget.'''

    def __init__(self, parent=None, viewer=viewer2D, shape=(600, 600), interactorStyle=None, title=""):
        '''Creates an instance of a `QCILDockableWidget` and inserts it in a `QDockWidget`. 
        Sets the title of the dock widget.'''
        super(QCILDockableWidget, self).__init__(parent)

        self.frame = QCILViewerWidget(parent, viewer, shape=shape, interactorStyle=interactorStyle)
        self.viewer = self.frame.viewer

        self.setWindowTitle(title)

        self.setWidget(self.frame)



import os
import unittest

import numpy as np
import os
import unittest

import h5py
import numpy as np
import vtk
from ccpi.viewer.utils.conversion import Converter
from ccpi.viewer.utils.io import cilviewerHDF5Reader
import yaml

from os import system


def calculate_target_downsample_shape(max_size, total_size, shape, acq=False):
    if not acq:
        xy_axes_magnification = np.power(max_size / total_size, 1 / 3)
        slice_per_chunk = int(1 / xy_axes_magnification)
    else:
        slice_per_chunk = 1
        xy_axes_magnification = np.power(max_size / total_size, 1 / 2)
    num_chunks = 1 + len([i for i in range(slice_per_chunk, shape[2], slice_per_chunk)])

    target_image_shape = (int(xy_axes_magnification * shape[0]), int(xy_axes_magnification * shape[1]), num_chunks)
    return target_image_shape


class TestCLIResample(unittest.TestCase):

    def setUp(self):
        # Generate random 3D array:
        bits = 16
        self.bytes_per_element = int(bits / 8)
        self.input_3D_array = np.random.randint(10, size=(5, 10, 6), dtype=eval(f"np.uint{bits}"))

        # write to HDF5: -----------
        self.hdf5_filename_3D = 'test_3D_data.h5'
        self.hdf5_yaml_filename = 'test_hdf5.yaml'
        with h5py.File(self.hdf5_filename_3D, 'w') as f:
            f.create_dataset('/entry1/tomo_entry/data/data', data=self.input_3D_array)

        self.hdf5_dict = {
            'input': {
                'file_name': self.hdf5_filename_3D,
                'dataset_name': '/entry1/tomo_entry/data/data'
            },
            'resample': {
                'target_size': 100e-6,
                'resample_z': False
            },
            'output': {
                'file_name': 'test_hdf5_out.hdf5',
                'format': 'hdf5'
            }
        }
        with open(self.hdf5_yaml_filename, 'w') as file:
            yaml.dump(self.hdf5_dict, file)

        # write to raw: -------------
        bytes_3D_array = bytes(self.input_3D_array)
        self.raw_filename_3D = 'test_3D_data.raw'
        self.raw_yaml_filename = 'test_raw.yaml'
        with open(self.raw_filename_3D, 'wb') as f:
            f.write(bytes_3D_array)
        self.raw_dict = {
            'input': {
                'file_name': self.raw_filename_3D,
                'shape': str(np.shape(self.input_3D_array)),
                'is_fortran': False,
                'is_big_endian': False,
                'typecode': str(self.input_3D_array.dtype)
            },
            'resample': {
                'target_size': 100e-6,
                'resample_z': False
            },
            'output': {
                'file_name': 'test_raw_out.hdf5',
                'format': 'hdf5'
            }
        }
        with open(self.raw_yaml_filename, 'w') as file:
            yaml.dump(self.raw_dict, file)

    def _test_resampling_acq_data(self, reader, target_size):
        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2] * self.bytes_per_element
        reader.Update()
        image = reader.GetOutputDataObject(0)
        extent = image.GetExtent()

        shape = calculate_target_downsample_shape(target_size, og_size, og_shape, acq=True)
        expected_size = shape[0] * shape[1] * shape[2]
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        resulting_size = resulting_shape[0] * \
            resulting_shape[1]*resulting_shape[2]
        # angle (z direction) is first index in numpy array, and in cil
        # but it is the last in vtk.
        resulting_z_shape = extent[5] + 1
        og_z_shape = np.shape(self.input_3D_array)[0]

        self.assertEqual(resulting_size, expected_size)
        self.assertEqual(resulting_z_shape, og_z_shape)

    def test_resample_with_yaml(self):
        dicts = [self.hdf5_dict, self.raw_dict]
        filenames = [self.hdf5_yaml_filename, self.raw_yaml_filename]
        subtest_labels = ['HDF5', 'raw']
        for i, dict in enumerate(dicts):
            with self.subTest(format=subtest_labels[i]):
                yaml_file = filenames[i]

            # Tests image with correct target size is generated by resample reader:
            system('resample -f {}'.format(yaml_file))

            reader = cilviewerHDF5Reader()
            reader.SetFileName(dict['output']['file_name'])
            target_size = int(dict['resample']['target_size'] * 1e6)
            self._test_resampling_acq_data(reader, target_size)

    def test_resample_command_line_hdf5(self):
        dict = self.hdf5_dict

        # Tests image with correct target size is generated by resample reader:
        input_file = dict['input']['file_name']
        target_size = dict['resample']['target_size']
        resample_z = dict['resample']['resample_z']
        out = dict['output']['file_name']
        out_format = dict['output']['format']

        # specific to hdf5:
        dset_name = dict['input']['dataset_name']

        command = f'resample -i {input_file} --dataset_name {dset_name} -o {out} -target_size {target_size} --resample_z {resample_z} --out_format {out_format}'

        if system(command) != 0:
            raise Exception("Error running test_resample_command_line_hdf5")

        reader = cilviewerHDF5Reader()
        reader.SetFileName(dict['output']['file_name'])

        target_size = int(target_size * 1e6)
        self._test_resampling_acq_data(reader, target_size)

    def test_resample_command_line_raw(self):
        dict = self.raw_dict

        # Tests image with correct target size is generated by resample reader:
        input_file = dict['input']['file_name']
        target_size = dict['resample']['target_size']
        resample_z = dict['resample']['resample_z']
        out = dict['output']['file_name']
        out_format = dict['output']['format']

        # specific to raw:
        shape = list(eval(dict['input']['shape']))
        shape = f"{shape[0]},{shape[1]},{shape[2]}"
        is_fortran = dict['input']['is_fortran']
        # is_fortran = False
        is_big_endian = dict['input']['is_big_endian']
        typecode = dict['input']['typecode']

        command = f'resample -i {input_file} --shape {shape} --is_fortran {is_fortran} --is_big_endian {is_big_endian} --typecode {typecode} -o {out} -target_size {target_size} --resample_z {resample_z} --out_format {out_format}'

        if system(command) != 0:
            raise Exception("Error running test_resample_command_line_raw")

        reader = cilviewerHDF5Reader()
        reader.SetFileName(dict['output']['file_name'])

        target_size = int(target_size * 1e6)
        self._test_resampling_acq_data(reader, target_size)

    def tearDown(self):
        files = [self.hdf5_filename_3D, self.hdf5_yaml_filename, self.raw_filename_3D, self.raw_yaml_filename]
        for f in files:
            os.remove(f)


if __name__ == '__main__':
    unittest.main()



import vtk
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from qtpy.QtCore import Qt, QEvent


class QCILRenderWindowInteractor(QVTKRenderWindowInteractor):
    '''
    A QVTKRenderWindowInteractor for Python and Qt. Uses a vtkGenericRenderWindowInteractor to handle the interactions.
    Use GetRenderWindow() to get the vtkRenderWindow. Create with the keyword stereo=1 in order to generate
    a stereo-capable window. Extends the QVTKRenderWindowInteractor to accept also ALT modifier.

    More info: https://docs.vtk.org/en/latest/api/python/vtkmodules/vtkmodules.qt.QVTKRenderWindowInteractor.html
    '''

    def __init__(self, parent=None, **kw):
        '''Constructor'''
        super(QCILRenderWindowInteractor, self).__init__(parent, **kw)
        self.__saveModifiers = self._QVTKRenderWindowInteractor__saveModifiers
        self.__saveX = self._QVTKRenderWindowInteractor__saveX
        self.__saveY = self._QVTKRenderWindowInteractor__saveY
        self.__saveButtons = self._QVTKRenderWindowInteractor__saveButtons
        self.__wheelDelta = self._QVTKRenderWindowInteractor__wheelDelta
        #print ("__saveModifiers  should be defined", self.__saveModifiers)

    def _GetAlt(self, ev):
        '''Get ALT key modifier'''
        ctrl = shift = alt = False

        if hasattr(ev, 'modifiers'):
            if ev.modifiers() & Qt.AltModifier:
                alt = True
        else:
            if self.__saveModifiers & Qt.AltModifier:
                alt = True

        return alt

    def enterEvent(self, ev):
        '''Overload of enterEvent from base class to get also the Alt modifier'''
        alt = self._GetAlt(ev)
        self._Iren.SetAltKey(alt)
        super().enterEvent(ev)

    def leaveEvent(self, ev):
        '''Overload of leaveEvent from base class to get also the Alt modifier'''
        alt = self._GetAlt(ev)
        self._Iren.SetAltKey(alt)
        super().leaveEvent(ev)

    def mousePressEvent(self, ev):
        '''Overload of mousePressEvent from base class to get also the Alt modifier'''
        alt = self._GetAlt(ev)
        self._Iren.SetAltKey(alt)
        super().mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        '''Overload of mousePressEvent from base class to get also the Alt modifier'''
        alt = self._GetAlt(ev)
        self._Iren.SetAltKey(alt)
        super().mouseReleaseEvent(ev)

    def mouseMoveEvent(self, ev):
        '''Overload of mouseMoveEvent from base class to get also the Alt modifier'''
        alt = self._GetAlt(ev)
        self._Iren.SetAltKey(alt)
        super().mouseMoveEvent(ev)



import vtk
import logging

logger = logging.getLogger(__name__)
# Define a new event type
SLIDER_EVENT = vtk.vtkCommand.UserEvent + 1


class SliceSliderRepresentation(vtk.vtkSliderRepresentation2D):
    """A slider representation for the slice selector slider on a 2D CILViewer

    Parameters
    -----------
    orientation: str, optional
        The orientation of the slider. Can be 'horizontal' or 'vertical'
    offset: float, optional
        The offset of the slider from the edge of the window. Default is 0.12

    """

    def __init__(self, orientation='horizontal', offset=0.12):
        self.tube_width = 0.004
        self.slider_length = 0.015
        self.slider_width = 0.015
        self.end_cap_length = 0.008
        self.end_cap_width = 0.02
        self.title_height = 0.02
        self.label_height = 0.02
        self.bar_color = 'Gray'
        cil_pink = [[el / 0xff for el in [0xe5, 0x06, 0x95]], [el / 0xff for el in [0xc9, 0x2c, 0x99]],
                    [el / 0xff for el in [0x99, 0x3d, 0xbb]], [el / 0xff for el in [0x51, 0x0c, 0x76]]]

        self.orientation = 'horizontal'
        self.offset = 0.12

        self.p1 = [self.offset, self.end_cap_width * 1.1]
        self.p2 = [1 - self.offset, self.end_cap_width * 1.1]

        self.title = None

        if orientation == 'vertical':
            self.offset = offset
            self.p1 = [self.end_cap_width * 1.1, self.offset]
            self.p2 = [self.end_cap_width * 1.1, 1 - self.offset]

        self.SetTitleText(self.title)

        self.GetPoint1Coordinate().SetCoordinateSystemToNormalizedDisplay()
        self.GetPoint1Coordinate().SetValue(self.p1[0], self.p1[1])
        self.GetPoint2Coordinate().SetCoordinateSystemToNormalizedDisplay()
        self.GetPoint2Coordinate().SetValue(self.p2[0], self.p2[1])

        self.SetTubeWidth(self.tube_width)
        self.SetSliderLength(self.slider_length)
        # slider_width = self.tube_width
        # slider.SetSliderWidth(slider_width)
        self.SetEndCapLength(self.end_cap_length)
        self.SetEndCapWidth(self.end_cap_width)
        self.SetTitleHeight(self.title_height)
        self.SetLabelHeight(self.label_height)

        # Set the colors of the slider components.
        # Change the color of the bar.
        self.GetTubeProperty().SetColor(vtk.vtkNamedColors().GetColor3d(self.bar_color))
        # Change the color of the ends of the bar.
        self.GetCapProperty().SetColor(cil_pink[2])
        # Change the color of the knob that slides.
        self.GetSliderProperty().SetColor(cil_pink[1])
        # Change the color of the knob when the mouse is held on it.
        self.GetSelectedProperty().SetColor(cil_pink[0])


class SliderCallback:
    '''
    Class to propagate the effects of interaction between the slider widget and the viewer 
    the slider is embedded into, and viceversa.
    
    Parameters
    -----------
    viewer : CILViewer2D the slider is embedded into
    slider_widget : the vtkSliderWidget that is embedded in the viewer
    '''

    def __init__(self, viewer, slider_widget):
        self.viewer = viewer
        self.slider_widget = slider_widget

    def __call__(self, caller, ev):
        '''Update the slice displayed by the viewer when the slider is moved
        
        Parameters
        -----------
        caller : the slider widget
        ev : the event that triggered the update
        '''
        slider_widget = caller
        value = slider_widget.GetRepresentation().GetValue()
        self.viewer.displaySlice(int(value))
        self.update_label(value)
        self.viewer.getInteractor().InvokeEvent(SLIDER_EVENT)

    def update_label(self, value):
        '''Update the text label on the slider. This is called by update_from_viewer
        
        Parameters
        -----------
        value : the value to be displayed on text label the slider
        '''
        rep = self.slider_widget.GetRepresentation()
        maxval = rep.GetMaximumValue()
        txt = "{}/{}".format(int(value), int(maxval))
        rep.SetLabelFormat(txt)

    def update_from_viewer(self, caller, ev):
        '''Update the slider widget from the viewer. This is called when the viewer changes the slice
        
        Parameters
        -----------
        caller : the interactor style
        ev : the event that triggered the update
        '''
        # The caller is the interactor style
        logger.info(f"Updating for event {ev}")
        value = caller.GetActiveSlice()
        self.slider_widget.GetRepresentation().SetValue(value)
        self.update_label(value)
        caller.GetRenderWindow().Render()

    def update_orientation(self, caller, ev):
        '''Update the slider widget when the orientation is changed

        Parameters
        -----------
        caller : the interactor style
        ev : the event that triggered the update
        '''
        logger.info(f"Updating orientation {ev}")
        value = caller.GetActiveSlice()
        dims = caller._viewer.img3D.GetDimensions()
        maxslice = dims[caller.GetSliceOrientation()] - 1
        sr = self.slider_widget.GetRepresentation()
        sr.SetMinimumValue(0)
        sr.SetMaximumValue(maxslice)
        self.update_from_viewer(caller, ev)
